/**************************************************************************
FILE          : Manager.c
PROGRAMMER    : Peter C. Chapin
LAST REVISION : 2012-11-11

This file contains a skeleton program to illustrate the basics of making a resource manager
under QNX. This code is pretty much the sample provided with the QNX documentation.

Please send comments or bug reports to

     Peter C. Chapin
     Computer Information Systems Department
     Vermont Technical College
     Williston, VT 05495
     PChapin@vtc.vsc.edu
**************************************************************************/

#include <errno.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/iofunc.h>
#include <sys/resmgr.h>
#include <sys/dispatch.h>

// The function that handles the _IO_READ and _IO_WRITE messages.
int io_read( resmgr_context_t *ctp, io_read_t *msg, RESMGR_OCB_T *ocb );
int io_write( resmgr_context_t *ctp, io_write_t *msg, RESMGR_OCB_T *ocb );

// Contents of my "file".
char *buffer = "Hello, World!\n";

// The following store the jump table used by the dispatcher.
static resmgr_connect_funcs_t connect_funcs;
static resmgr_io_funcs_t      io_funcs;

// The attributes of the pathname.
static iofunc_attr_t attr;


//
// Main Program
// 

int main( int argc, char **argv )
{
    resmgr_attr_t       resmgr_attr;  // Resource manager attributes.
    dispatch_t         *dpp;          // Dispatch handle.
    dispatch_context_t *ctp;          // Where message buffer, etc, is stored.
    int                 id;

    // Initialize dispatch interface.
    if( ( dpp = dispatch_create() ) == NULL ) {
        fprintf( stderr, "%s: Unable to allocate dispatch handle.\n", argv[0] );
        return EXIT_FAILURE;
    }

    // Initialize resource manager attributes.
    memset( &resmgr_attr, 0, sizeof(resmgr_attr) );
    resmgr_attr.nparts_max   = 1;    // Number of IOV segments for replies.
    resmgr_attr.msg_max_size = 2048; // Minimum RX buffer size.

    // Initialize functions for handling messages with defaults.
    iofunc_func_init(
        _RESMGR_CONNECT_NFUNCS, &connect_funcs,
        _RESMGR_IO_NFUNCS,      &io_funcs
    );
    // Override defaults for read and write messages.
    io_funcs.read  = io_read;
    io_funcs.write = io_write;

    // Initialize attribute structure used by the device. The initialization function takes file
    // type, permissions, owner and group IDs. Afterwards Set the size of the "file" to the
    // number of characters in my message.
    //
    iofunc_attr_init( &attr, S_IFNAM | 0666, 0, 0 );
    attr.nbytes = strlen( buffer );

    // Attach our device name. This registers the given name with the process manager and makes
    // our name visible in the file system name space.
    //
    id = resmgr_attach(
        dpp,            // Dispatch handle.
        &resmgr_attr,   // Resource manager attributes.
        "/root/sample", // Device name.
        _FTYPE_ANY,     // Open type. This defines what open modes we support.
        0,              // Flags. Controls how much pathname space we handle.
                        //   Zero means: we only handle the exact name registered.
        &connect_funcs, // Connect routines.
        &io_funcs,      // I/O routines.
        &attr           // Attributes of the name in the file system namespace.
    );
    if( id == -1 ) {
        fprintf( stderr, "%s: Unable to attach name.\n", argv[0] );
        return EXIT_FAILURE;
    }

    // Allocate a context structure.
    ctp = dispatch_context_alloc( dpp );

    // Start the resource manager message loop.
    while( 1 ) {
        if( ( ctp = dispatch_block(ctp) ) == NULL ) {
            fprintf( stderr, "%s: Error in dispatch_block().\n", argv[0] );
            return EXIT_FAILURE;
        }
        dispatch_handler( ctp );
    }
   
    return 0;
}


//
// The following function services the _IO_READ message.
// 

int io_read ( resmgr_context_t *ctp, io_read_t *msg, RESMGR_OCB_T *ocb )
{
    int nleft;   // Number of bytes left in the "file".
    int nbytes;  // Number of bytes we actually send to the client.
    int nparts;  // Number of parts in the outgoing IOV (either 0 or 1).
    int status;
   
    if( ( status = iofunc_read_verify( ctp, msg, ocb, NULL ) ) != EOK )
        return status;
           
    if( ( msg->i.xtype & _IO_XTYPE_MASK ) != _IO_XTYPE_NONE )
        return ENOSYS;
   
    // On all reads (first and subsequent) calculate how many bytes we can return to the client
    // based upon the number of bytes available (nleft) and the client's buffer size.
    //
    nleft  = ocb->attr->nbytes - ocb->offset;
    nbytes = min( msg->i.nbytes, nleft );
   
    if( nbytes > 0 ) {
        // Set up the return data IOV.
        SETIOV( ctp->iov, buffer + ocb->offset, nbytes );
	  
        // set up the number of bytes ( returned by client's read() ).
        _IO_SET_READ_NBYTES( ctp, nbytes );
	  
        // Advance the offset by the number of bytes returned to the client.
        ocb->offset += nbytes;
       
        nparts = 1;
    }
    else {

        // They've asked for zero bytes or they've already previously read everything
        _IO_SET_READ_NBYTES( ctp, 0 );
        nparts = 0;
    }
   
    // Mark the access time as invalid (we just accessed it).
    if( msg->i.nbytes > 0 )
        ocb->attr->flags |= IOFUNC_ATTR_ATIME;
   
    return _RESMGR_NPARTS( nparts );
}


//
// The following function services the _IO_WRITE message.
// 

int io_write( resmgr_context_t *ctp, io_write_t *msg, RESMGR_OCB_T *ocb )
{
    int   status;
    char *buf;

    if( ( status = iofunc_write_verify(ctp, msg, ocb, NULL ) ) != EOK )
        return status;

    if( ( msg->i.xtype & _IO_XTYPE_MASK) != _IO_XTYPE_NONE )
        return ENOSYS;

    // Set up the number of bytes written (returned by client's write()).
    _IO_SET_WRITE_NBYTES( ctp, msg->i.nbytes );
   
    buf = (char *)malloc( msg->i.nbytes + 1 );
    if( buf == NULL ) return ENOMEM;
  
    // Reread the data from the sender's message buffer. We're not assuming that all of the data
    // fit into the resource manager library's receive buffer.
    //
    resmgr_msgread( ctp, buf, msg->i.nbytes, sizeof(msg->i) );
    buf[msg->i.nbytes] = '\0';
    printf( "Received %d bytes = '%s'\n", msg->i.nbytes, buf );
    free( buf );

    if( msg->i.nbytes > 0 )
        ocb->attr->flags |= IOFUNC_ATTR_MTIME | IOFUNC_ATTR_CTIME;

    return _RESMGR_NPARTS( 0 );
}
