/****************************************************************************
FILE          : MsgTest.c
LAST REVISED  : 2012-11-11
SUBJECT       : Program to demonstrate QNX's message passing API.

This program illustrates how to set up a message channel, attach to it, and pass messages over
it. In this case the thread that acts as the client is in the same process as the thread that
acts as the server. However, by changing the node ID and the process ID in the ConnectAttach
call, threads in other processes or on other nodes can send messages just as easily.

     Peter C. Chapin
     Computer Information Systems Department
     Vermont Technical College
     Williston, VT 05495
     PChapin@vtc.vsc.edu
****************************************************************************/

#include <stdio.h>
#include <pthread.h>
#include <sys/netmgr.h>
#include <sys/neutrino.h>

//
// The following function acts as a simple client.
//
void *thread_function( void *id )
{
    int  channel_id = *(int *)id;
    int  connection_id;
    char reply_buffer[128];  // Holds the server's reply.
    int  status;             // Status code from the server.

    // Establish a connection with the server.
    connection_id = ConnectAttach( ND_LOCAL_NODE, 0, channel_id, _NTO_SIDE_CHANNEL, 0 );

    // Send a message and wait for the reply. Display result.
    status = MsgSend( connection_id, "Hello, World!", 14, reply_buffer, 128 );
    printf( "Client: Got reply: %s, status = %d\n", reply_buffer, status );

    // Break the connection with the server.
    ConnectDetach( connection_id );
    return NULL;
}


//
// The main program allocates the resources needed and acts as the server.
//
int main( void )
{
    int   channel_id;     // Process relative.
    pthread_t thread_id;  // Client thread ID.
    char  buffer[128];    // Hold's clients' request.
    int   receive_id;     // Used to identify the request (used for reply).

    // Create a channel.
    channel_id = ChannelCreate( 0 );
    printf( "Server: Created channel %d\n", channel_id );

    // Create the client thread and tell it about the channel.
    pthread_create( &thread_id, NULL, thread_function, &channel_id );

    // Get a client request, process it, and send back a reply.
    receive_id = MsgReceive( channel_id, buffer, 128, NULL );
    printf( "Server: Got %s\n", buffer );
    MsgReply( receive_id, 0, "OK", 3 );

    // Wait for the client thread to terminate.
    pthread_join( thread_id, NULL );

    // Clean up the channel.
    ChannelDestroy( channel_id );  
    return 0;
}
