/**************************************************************************
FILE         : SerialDriver.c
SUBJECT      : QNX serial driver demonstration.
LAST REVISED : 2008-11-09
PROGRAMMER   : (C) Copyright 2008 by Peter C. Chapin

This file contains a QNX resource manager for the serial port. This version is not very finished
and, in fact, rather stinks. However, it is a start. It is largely based on the sample code
provided with the QNX documentation.

Please send comments or bug reports to

     Peter C. Chapin
     Computer Information Systems Department
     Vermont Technical College
     Randolph Center, VT 05061
     Peter.Chapin@vtc.vsc.edu
**************************************************************************/

// Standard C
#include <errno.h>
#include <stdio.h>
#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

// POSIX
#include <unistd.h>
#include <pthread.h>
#include <semaphore.h>

// QNX
#include <sys/iofunc.h>
#include <sys/resmgr.h>
#include <sys/dispatch.h>
#include <sys/neutrino.h>
#include <hw/inout.h>

// The functions that handle the _IO_READ and _IO_WRITE messages.
int io_read(resmgr_context_t *ctp, io_read_t *msg, RESMGR_OCB_T *ocb);
// int io_write(resmgr_context_t *ctp, io_write_t *msg, RESMGR_OCB_T *ocb);

// The following store the jump table used by the dispatcher.
static resmgr_connect_funcs_t connect_funcs;
static resmgr_io_funcs_t      io_funcs;

// The attributes of the pathname.
static iofunc_attr_t attr;

// Nice names for UART registers.
#define RX                0  // Read only.
#define TX                0  // Write only.
#define DL0               0  // Divisor latch (LSB)
#define DL1               1  // Divisor latch (MSB)
#define INTERRUPT_ENABLE  1
#define INTERRUPT_ID      2  // Read only.
#define FIFO_CONTROL      2  // Write only.
#define LINE_CONTROL      3
#define MODEM_CONTROL     4
#define LINE_STATUS       5
#define MODEM_STATUS      6

// The COM1 port address (this program only handles COM1).
static const int port_address = 0x3F8;
#define IRQ_NUMBER 4

// Returned by the ISR (below) to indicate how interrupts should be handled.
static struct sigevent event;

// Material required to support the RX data.
#define RXBUFFER_SIZE 2048
static uint8_t rx_buffer[RXBUFFER_SIZE];
static int     rx_in    = 0;  // Index of next in position.
static int     rx_out   = 0;  // Index of next out position.
static int     rx_count = 0;  // Number of bytes in the buffer.
static pthread_mutex_t rx_mutex       = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t  data_available = PTHREAD_COND_INITIALIZER;


// The following function initializes the serial port. To keep my life simple, I will hard code
// various settings here. In real life this would probably all be provided via the command line.
//
void initialize_port()
{
    // Set N81 with DLAB enabled.
    out8(port_address + LINE_CONTROL, 0x83);

    // Set the divisor latch to produce 9600 baud.
    out8(port_address + DL0, 12);
    out8(port_address + DL1, 0);

    // Set N81 with DLAB disabled.
    out8(port_address + LINE_CONTROL, 0x03);

    // Enable interrupts (just RX for now).
    out8(port_address + INTERRUPT_ENABLE, 0x01);
}


// The following function is the actual ISR. It simply returns an indication saying that
// InterruptWait should be unblocked. Most of the real work of dealing with the interrupt is
// done in the hardware_thread.
// 
const struct sigevent *isr(void *arg, int ID)
{
  // Look at the hardware to see if this interrupt really happened. (Are spurious interrupts
  // possible in this configuration?)
  //
  uint8_t check_interrupt = in8(port_address + INTERRUPT_ID);
  
  // Ignore everything except for the RX data interrupt.
  if ((check_interrupt & 0x0F) != 4) return NULL;
   
  // We won't read the UART here. Let the UART thread deal with that.
  InterruptMask(IRQ_NUMBER, -1);
 
  return &event;
}


// The following function does the grunt work of servicing an interrupt. It first programs the
// hardware and then waits for an interrupt to occur. When an interrupt happens it deals with
// the hardware and with the other threads that are clamoring for it. This thread thus serves as
// a mediator between the hardware and the rest of the system.
// 
void *hardware_thread(void *arg)
{
    int     interrupt_id;
    uint8_t cause;
    uint8_t raw_byte;
  
    // Used for modifying the scheduling policy for this thread.
    int policy;
    struct sched_param parameters;

    // This thread needs access to I/O devices. (Only root can do this).
    ThreadCtl(_NTO_TCTL_IO, 0);
   
    // Initialize the segevent object for use by the ISR.
    SIGEV_INTR_INIT(&event);
   
    // Set up the hardware (this includes enabling interrupts).
    initialize_port();
   
    // Inform QNX of our ISR.
    interrupt_id = InterruptAttach(IRQ_NUMBER, isr, NULL, 0, 0);
    if (interrupt_id == -1) {
        fprintf(stderr, "Failed to attach the interrupt service routine!\n");
        return NULL;
    }
   
    // // Do I want to boost this thread's priority? Probably. Let's do it!
    // pthread_getschedparam(pthread_self(), &policy, &parameters);
    // policy = SCHED_FIFO;
    // parameters.sched_priority += 5;
    // pthread_setschedparam(pthread_self(), policy, &parameters);
  
    // Assert RTS to tell the other side that we are ready to receive.
    out8(port_address + MODEM_CONTROL, in8(port_address + MODEM_CONTROL) | 0x02);
  
    while (1) {
        InterruptWait(0, NULL);
     
        // Figure out why this interrupt occurred. Currently the ISR only allows interrupts due
        // to RX data to be handled. However, a future version of this manager might deal with
        // other kinds of interrupts.
        //
        cause = in8(port_address + INTERRUPT_ID);
        cause = (cause & 0x0E) >> 1;
    
        switch (cause) {
        case 2: // RX data available.
            // Get the data (releasing interrupt request).
            raw_byte = in8(port_address + RX);
            InterruptUnmask(IRQ_NUMBER, -1);

            pthread_mutex_lock(&rx_mutex);
            rx_buffer[rx_in] = raw_byte;
            ++rx_in;
            if (rx_in >= RXBUFFER_SIZE) rx_in = 0;
            ++rx_count;
        	
            // Drop RTS if the buffer is more than half full for flow control.
            if (rx_count > RXBUFFER_SIZE/2) {
                out8(port_address + MODEM_CONTROL, in8(port_address + MODEM_CONTROL) & 0xFD);
            }
        	
            // Signal the data available condition if this is the first byte.
            if (rx_count == 1) pthread_cond_signal(&data_available);
            pthread_mutex_unlock(&rx_mutex);
            break;
        }
    }
    
    InterruptDetach(interrupt_id);
   
    // This function never returns.
    return NULL;
}


// ------------
// Main Program
// ------------

int main(int argc, char **argv)
{
    // Declare variables we'll be using.
    resmgr_attr_t       resmgr_attr;  // Resource manager attributes.
    dispatch_t         *dpp;          // Dispatch thingy.
    dispatch_context_t *ctp;          // Context thingy.
    int                 id;
    pthread_t           thread_id;    // ID of the hardware thread.

    // Start the hardware control thread.
    pthread_create(&thread_id, NULL, hardware_thread, NULL);
   
    // Initialize dispatch interface.
    if((dpp = dispatch_create()) == NULL) {
        fprintf(stderr, "%s: Unable to allocate dispatch handle.\n", argv[0]);
        return EXIT_FAILURE;
    }

    // Initialize resource manager attributes.
    memset(&resmgr_attr, 0, sizeof(resmgr_attr));
    resmgr_attr.nparts_max   = 1;
    resmgr_attr.msg_max_size = 2048;

    // Initialize functions for handling messages. Specify my io functions and override the
    // default.
    //
    iofunc_func_init(_RESMGR_CONNECT_NFUNCS, &connect_funcs, _RESMGR_IO_NFUNCS, &io_funcs);
    io_funcs.read = io_read;
    // io_funcs.write = io_write;

    // Initialize attribute structure used by the device.
    iofunc_attr_init(&attr, S_IFNAM | 0666, 0, 0);
    attr.nbytes = 0;

    // Attach our device name.
    id = resmgr_attach(
            dpp,            // Dispatch handle.
            &resmgr_attr,   // Resource manager attrs.
            "/root/serial", // Device name.
            _FTYPE_ANY,     // Open type.
            0,              // Flags.
            &connect_funcs, // Connect routines.
            &io_funcs,      // I/O routines.
            &attr
    );
    if (id == -1) {
        fprintf(stderr, "%s: Unable to attach name.\n", argv[0]);
        return EXIT_FAILURE;
    }

    // Allocate a context structure.
    ctp = dispatch_context_alloc(dpp);

    // Start the resource manager message loop.
    while(1) {
        if ((ctp = dispatch_block(ctp)) == NULL) {
            fprintf(stderr, "block error\n");
            return EXIT_FAILURE;
        }
        dispatch_handler(ctp);
    }
     
    return 0;
}


// The following function services the _IO_READ message.
// 
int io_read(resmgr_context_t *ctp, io_read_t *msg, RESMGR_OCB_T *ocb)
{
    int i;
    int nbytes;
    int nparts;
    int status;
    unsigned char buffer[RXBUFFER_SIZE];
 
    // This thread needs access to I/O devices as well (to raise RTS). Can this operation be
    // done in main()?
    //
    ThreadCtl(_NTO_TCTL_IO, 0);

    // Make sure this is really a ready operation.  
    if ((status = iofunc_read_verify(ctp, msg, ocb, NULL)) != EOK)
        return status;
 
    // We don't support extended message types.          
    if ((msg->i.xtype & _IO_XTYPE_MASK) != _IO_XTYPE_NONE)
        return ENOSYS;

    // Protect ourselves from the hardware thread.
    pthread_mutex_lock(&rx_mutex);
  
    // If nothing here, wait until there is.
    if (rx_count == 0)
        pthread_cond_wait(&data_available, &rx_mutex);
  
    // Compute the amount of data to send and copy it into a temporary holding area.
    nbytes = min(msg->i.nbytes, rx_count);
    for (i = 0; i < nbytes; ++i) {
        buffer[i] = rx_buffer[rx_out++];
        if (rx_out >= RXBUFFER_SIZE) rx_out = 0;
    }
    rx_count -= nbytes;

    // Raise RTS if the buffer is 1/4 full (or less) for flow control.
    if (rx_count <= RXBUFFER_SIZE/4) {
        out8(port_address + MODEM_CONTROL, in8(port_address + MODEM_CONTROL) | 0x02);
    }
        	
    pthread_mutex_unlock(&rx_mutex);

    if (nbytes > 0) {
        // Set up the return data IOV.
        SETIOV(ctp->iov, buffer, nbytes);
	  
        // set up the number of bytes (returned by client's read()).
        _IO_SET_READ_NBYTES(ctp, nbytes);
       
        nparts = 1;
    }
    else {
        // This occurs only if they requested zero bytes.
        _IO_SET_READ_NBYTES(ctp, 0);
	          
        nparts = 0;
    }
   
    // Mark the access time as invalid (we just accessed it).
    if (msg->i.nbytes > 0)
        ocb->attr->flags |= IOFUNC_ATTR_ATIME;
   
    return _RESMGR_NPARTS(nparts);
}



// The following function services the _IO_WRITE message.
// 
// int io_write(resmgr_context_t *ctp, io_write_t *msg, RESMGR_OCB_T *ocb)
// {
//     return _RESMGR_NPARTS(0);
// }
